# -*- coding: utf-8 -*-
# noqa: D205,D400
"""
Indicators module
=================

Indicators are the main tool xclim provides to compute climate indices. In contrast
to the function defined in `xclim.indices`, Indicators add a layer of health checks
and metadata handling. Indicator objects are split into realms : atmos, land and
seaIce.

The module also defines additional virtual modules : `icclim` and `anuclim`. For the moment, they hold indices
instead of indicators, but in the future they'll be converted to indicators.
"""
from functools import wraps
from types import ModuleType
from typing import Callable


def build_module(
    name: str,
    objs: dict,
    doc: str = "",
    source: ModuleType = None,
    mode: str = "ignore",
):
    """Create a module from imported objects.

    Parameters
    ----------
    name : str
      New module name.
    objs : dict
      Dictionary of the objects (or their name) to import into the module,
      keyed by the name they will take in the created module.
    doc : str
      Docstring of the new module.
    source : Module object
      Module where objects are defined if not explicitly given.
    mode : {'raise', 'warn', 'ignore'}
      How to deal with missing objects.


    Returns
    -------
    ModuleType
      A module built from a list of objects' name.

    """
    import logging
    import sys
    import types
    import warnings

    try:
        out = types.ModuleType(name, doc)
    except TypeError:
        msg = "Module '{}' is not properly formatted".format(name)
        raise TypeError(msg)

    for key, obj in objs.items():
        if isinstance(obj, str) and source is not None:
            module_mappings = getattr(source, obj, None)
        else:
            module_mappings = obj

        if module_mappings is None:
            msg = "{} has not been implemented.".format(obj)
            if mode == "ignore":
                logging.info(msg)
            elif mode == "warn":
                warnings.warn(msg)
            elif mode == "raise":
                raise NotImplementedError(msg)
            else:
                msg = "{} is not a valid missing object behaviour".format(mode)
                raise AttributeError(msg)

        else:
            out.__dict__[key] = module_mappings
            try:
                module_mappings.__module__ = name
            except AttributeError:
                msg = "{} is not a function".format(module_mappings)
                raise AttributeError(msg)

    sys.modules[name] = out
    return out


def __build_icclim(mode: str = "warn"):
    from xclim import indices
    from xclim.core.utils import wrapped_partial

    #  ['SD', 'SD1', 'SD5cm', 'SD50cm',
    # TODO : Complete mappings for ICCLIM indices
    mapping = {
        "TG": indices.tg_mean,
        "TX": indices.tx_mean,
        "TN": indices.tn_mean,
        "TG90p": indices.tg90p,
        "TG10p": indices.tg10p,
        "TGx": indices.tg_max,
        "TGn": indices.tg_min,
        "TX90p": indices.tx90p,
        "TX10p": indices.tx10p,
        "TXx": indices.tx_max,
        "TXn": indices.tx_min,
        "TN90p": indices.tn90p,
        "TN10p": indices.tn10p,
        "TNx": indices.tn_max,
        "TNn": indices.tn_min,
        "CSDI": indices.cold_spell_duration_index,
        "SU": indices.tx_days_above,
        "CSU": indices.maximum_consecutive_tx_days,
        "TR": indices.tropical_nights,
        "GD4": wrapped_partial(indices.growing_degree_days, thresh="4 degC"),
        "FD": indices.frost_days,
        "CFD": wrapped_partial(indices.maximum_consecutive_frost_days, thresh="0 degC"),
        "GSL": indices.growing_season_length,
        "ID": indices.ice_days,
        "HD17": wrapped_partial(indices.heating_degree_days, thresh="17 degC"),
        "CDD": indices.maximum_consecutive_dry_days,
        "CWD": indices.maximum_consecutive_wet_days,
        "PRCPTOT": indices.precip_accumulation,
        "RR1": indices.wetdays,
        "SDII": wrapped_partial(indices.daily_pr_intensity, thresh="1 mm/day"),
        "ETR": indices.extreme_temperature_range,
        "DTR": indices.daily_temperature_range,
        "vDTR": indices.daily_temperature_range_variability,
        "R10mm": wrapped_partial(indices.wetdays, thresh="10 mm/day"),
        "R20mm": wrapped_partial(indices.wetdays, thresh="20 mm/day"),
        "RX1day": indices.max_1day_precipitation_amount,
        "RX5day": wrapped_partial(indices.max_n_day_precipitation_amount, window=5),
        "WSDI": indices.warm_spell_duration_index,
        "R75p": wrapped_partial(indices.days_over_precip_thresh, thresh="1 mm/day"),
        "R95p": wrapped_partial(indices.days_over_precip_thresh, thresh="1 mm/day"),
        "R99p": wrapped_partial(indices.days_over_precip_thresh, thresh="1 mm/day"),
        "R75pTOT": wrapped_partial(
            indices.fraction_over_precip_thresh, thresh="1 mm/day"
        ),
        "R95pTOT": wrapped_partial(
            indices.fraction_over_precip_thresh, thresh="1 mm/day"
        ),
        "R99pTOT": wrapped_partial(
            indices.fraction_over_precip_thresh, thresh="1 mm/day"
        ),
        # 'SD': None,
        # 'SD1': None,
        # 'SD5cm': None,
        # 'SD50cm': None,
    }

    mod = build_module(
        "xclim.icclim",
        mapping,
        doc="""
            ==============
            ICCLIM indices
            ==============
            The European Climate Assessment & Dataset project (`ECAD`_) defines
            a set of 26 core climate indices. Those have been made accessible
            directly in xclim through their ECAD name for compatibility. However,
            the methods in this module are only wrappers around the corresponding
            methods of  `xclim.indices`. Note that none of the checks performed by
            the `xclim.utils.Indicator` class (like with `xclim.atmos` indicators)
            are performed in this module.

            .. _ECAD: https://www.ecad.eu/
            """,
        mode=mode,
    )
    return mod


def ensure_annual(func: Callable) -> Callable:
    """Ensure that supplied frequency keyword denotes annual time step."""

    @wraps(func)
    def _wrapper(*args, **kwargs):
        if "freq" not in kwargs:
            raise ValueError(
                "Frequency must be provided as a keyword argument (freq='Y[S]' or freq='A[S][-month]')"
            )
        freq = kwargs["freq"]
        if freq[0] not in ["Y", "A"]:
            raise ValueError("Frequency must be annual ('Y[S]' or 'A[S][-month]')")
        return func(*args, **kwargs)

    return _wrapper


def __build_anuclim(mode: str = "warn"):
    from xclim import indices
    from xclim.core.utils import wrapped_partial

    mapping = {
        "P1_AnnMeanTemp": ensure_annual(indices.tg_mean),
        "P2_MeanDiurnalRange": ensure_annual(indices.daily_temperature_range),
        "P3_Isothermality": ensure_annual(indices.isothermality),
        "P4_TempSeasonality": indices.temperature_seasonality,
        "P5_MaxTempWarmestPeriod": ensure_annual(indices.tx_max),
        "P6_MinTempColdestPeriod": ensure_annual(indices.tn_min),
        "P7_TempAnnualRange": ensure_annual(indices.extreme_temperature_range),
        "P8_MeanTempWettestQuarter": ensure_annual(
            wrapped_partial(indices.tg_mean_wetdry_quarter, op="wettest")
        ),
        "P9_MeanTempDriestQuarter": ensure_annual(
            wrapped_partial(indices.tg_mean_wetdry_quarter, op="driest")
        ),
        "P10_MeanTempWarmestQuarter": ensure_annual(
            wrapped_partial(indices.tg_mean_warmcold_quarter, op="warmest")
        ),
        "P11_MeanTempColdestQuarter": ensure_annual(
            wrapped_partial(indices.tg_mean_warmcold_quarter, op="coldest")
        ),
        "P12_AnnualPrecip": ensure_annual(indices.prcptot),
        "P13_PrecipWettestPeriod": ensure_annual(
            wrapped_partial(indices.prcptot_wetdry_period, op="wettest")
        ),
        "P14_PrecipDriestPeriod": ensure_annual(
            wrapped_partial(indices.prcptot_wetdry_period, op="driest")
        ),
        "P15_PrecipSeasonality": indices.precip_seasonality,
        "P16_PrecipWettestQuarter": ensure_annual(
            wrapped_partial(indices.prcptot_wetdry_quarter, op="wettest")
        ),
        "P17_PrecipDriestQuarter": ensure_annual(
            wrapped_partial(indices.prcptot_wetdry_quarter, op="driest")
        ),
        "P18_PrecipWarmestQuarter": ensure_annual(
            wrapped_partial(indices.prcptot_warmcold_quarter, op="warmest")
        ),
        "P19_PrecipColdestQuarter": ensure_annual(
            wrapped_partial(indices.prcptot_warmcold_quarter, op="coldest")
        ),
    }

    mod = build_module(
        "xclim.anuclim",
        mapping,
        doc="""
                ==============
                ANUCLIM indices
                ==============

                The ANUCLIM (v6.1) software package' BIOCLIM sub-module produces a set of Bioclimatic
                parameters derived values of temperature and precipitation. The methods in this module
                are wrappers around a subset of corresponding methods of `xclim.indices`. Note that none
                of the checks performed by the `xclim.utils.Indicator` class (like with `xclim.atmos`
                indicators) are performed in this module.

                Futhermore, according to the ANUCLIM user-guide https://fennerschool.anu.edu.au/files/anuclim61.pdf (ch. 6),
                input values should be at a weekly (or monthly) frequency.  However, the xclim.indices
                implementation here will calculate the result with input data of any frequency.

                .. _ANUCLIM: https://fennerschool.anu.edu.au/files/anuclim61.pdf (ch. 6)
                """,
        mode=mode,
    )
    return mod


ICCLIM = icclim = __build_icclim("ignore")
anuclim = __build_anuclim()



"""Locales and language support module."""
from xclim.core.formatting import default_formatter
from xclim.core.locales import TRANSLATABLE_ATTRS, get_best_locale, get_local_dict


def generate_local_dict(locale: str, init_english: bool = False):
    """Generate a dictionary with keys for each indicators and translatable attributes.

    Parameters
    ----------
    locale : str
        Locale in the IETF format
    init_english : bool
        If True, fills the initial dictionary with the english versions of the attributes.
        Defaults to False.
    """
    from xclim.core.indicator import registry

    best_locale = get_best_locale(locale)
    if best_locale is not None:
        locname, attrs = get_local_dict(best_locale)
        for ind_name in attrs.copy().keys():
            if ind_name != "attrs_mapping" and ind_name not in registry:
                attrs.pop(ind_name)
    else:
        attrs = {}

    attrs_mapping = attrs.setdefault("attrs_mapping", {})
    attrs_mapping.setdefault("modifiers", [""])
    for key, value in default_formatter.mapping.items():
        attrs_mapping.setdefault(key, [value[0]])

    eng_attr = ""
    for ind_name, indicator in registry.items():
        ind_attrs = attrs.setdefault(ind_name, {})
        for translatable_attr in set(TRANSLATABLE_ATTRS).difference(
            set(indicator._cf_names)
        ):
            if init_english:
                eng_attr = getattr(indicator, translatable_attr)
                if not isinstance(eng_attr, str):
                    eng_attr = ""
            ind_attrs.setdefault(f"{translatable_attr}", eng_attr)

        for var_attrs in indicator.cf_attrs:
            # In the case of single output, put var attrs in main dict
            if len(indicator.cf_attrs) > 1:
                ind_attrs = attrs.setdefault(f"{ind_name}.{var_attrs['var_name']}", {})

            for translatable_attr in set(TRANSLATABLE_ATTRS).intersection(
                set(indicator._cf_names)
            ):
                if init_english:
                    eng_attr = var_attrs.get(translatable_attr)
                    if not isinstance(eng_attr, str):
                        eng_attr = ""
                ind_attrs.setdefault(f"{translatable_attr}", eng_attr)
    return attrs



# -*- coding: utf-8 -*-
# noqa: D205,D400
"""
Generic indices submodule
=========================

Helper functions for common generic actions done in the computation of indices.
"""
from typing import Union

import numpy as np
import xarray as xr

__all__ = [
    "select_time",
    "select_resample_op",
    "doymax",
    "doymin",
    "default_freq",
    "threshold_count",
    "get_daily_events",
    "daily_downsampler",
]


def select_time(da: xr.DataArray, **indexer):
    """Select entries according to a time period.

    Parameters
    ----------
    da : xr.DataArray
      Input data.
    **indexer : {dim: indexer, }, optional
      Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
      month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
      considered.

    Returns
    -------
    xr.DataArray
      Selected input values.
    """
    if not indexer:
        selected = da
    else:
        key, val = indexer.popitem()
        time_att = getattr(da.time.dt, key)
        selected = da.sel(time=time_att.isin(val)).dropna(dim="time")

    return selected


def select_resample_op(da: xr.DataArray, op: str, freq: str = "YS", **indexer):
    """Apply operation over each period that is part of the index selection.

    Parameters
    ----------
    da : xr.DataArray
      Input data.
    op : str {'min', 'max', 'mean', 'std', 'var', 'count', 'sum', 'argmax', 'argmin'} or func
      Reduce operation. Can either be a DataArray method or a function that can be applied to a DataArray.
    freq : str
      Resampling frequency defining the periods
      defined in http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling.
    **indexer : {dim: indexer, }, optional
      Time attribute and values over which to subset the array. For example, use season='DJF' to select winter values,
      month=1 to select January, or month=[6,7,8] to select summer months. If not indexer is given, all values are
      considered.

    Returns
    -------
    xarray.DataArray
      The maximum value for each period.
    """
    da = select_time(da, **indexer)
    r = da.resample(time=freq, keep_attrs=True)
    if isinstance(op, str):
        return getattr(r, op)(dim="time", keep_attrs=True)

    return r.map(op)


def doymax(da: xr.DataArray):
    """Return the day of year of the maximum value."""
    i = da.argmax(dim="time")
    out = da.time.dt.dayofyear[i]
    out.attrs["units"] = ""
    return out


def doymin(da: xr.DataArray):
    """Return the day of year of the minimum value."""
    i = da.argmin(dim="time")
    out = da.time.dt.dayofyear[i]
    out.attrs["units"] = ""
    return out


def default_freq(**indexer):
    """Return the default frequency."""
    freq = "AS-JAN"
    if indexer:
        group, value = indexer.popitem()
        if "DJF" in value:
            freq = "AS-DEC"
        if group == "month" and sorted(value) != value:
            raise NotImplementedError

    return freq


binary_ops = {">": "gt", "<": "lt", ">=": "ge", "<=": "le", "==": "eq", "!=": "ne"}


def get_op(op: str):
    """Get python's comparing function according to its name of representation.

    Accepted op string are keys and values of xclim.indices.generic.binary_ops.
    """
    if op in binary_ops:
        op = binary_ops[op]
    elif op in binary_ops.values():
        pass
    else:
        raise ValueError(f"Operation `{op}` not recognized.")
    return xr.core.ops.get_op(op)


def compare(da: xr.DataArray, op: str, thresh: Union[float, int]) -> xr.DataArray:
    """Compare a dataArray to a threshold using given operator.

    Parameters
    ----------
    da : xr.DataArray
      Input data.
    op : {">", "<", ">=", "<=", "gt", "lt", "ge", "le"}
      Logical operator {>, <, >=, <=, gt, lt, ge, le }. e.g. arr > thresh.
    thresh : Union[float, int]
      Threshold value.

    Returns
    -------
    xr.DataArray
        Boolean mask of the comparison.
    """
    func = getattr(da, "_binary_op")(get_op(op))
    return func(da, thresh)


def threshold_count(
    da: xr.DataArray, op: str, thresh: Union[float, int], freq: str
) -> xr.DataArray:
    """Count number of days where value is above or below threshold.

    Parameters
    ----------
    da : xr.DataArray
      Input data.
    op : {">", "<", ">=", "<=", "gt", "lt", "ge", "le"}
      Logical operator {>, <, >=, <=, gt, lt, ge, le }. e.g. arr > thresh.
    thresh : Union[float, int]
      Threshold value.
    freq : str
      Resampling frequency defining the periods
      defined in http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling.

    Returns
    -------
    xr.DataArray
      The number of days meeting the constraints for each period.
    """
    c = compare(da, op, thresh) * 1
    return c.resample(time=freq).sum(dim="time")


def domain_count(da: xr.DataArray, low: float, high: float, freq: str) -> xr.DataArray:
    """Count number of days where value is within low and high thresholds.

    A value is counted if it is larger than `low`, and smaller or equal to `high`, i.e. in `]low, high]`.

    Parameters
    ----------
    da : xr.DataArray
      Input data.
    low : float
      Minimum threshold value.
    high : float
      Maximum threshold value.
    freq : str
      Resampling frequency defining the periods
      defined in http://pandas.pydata.org/pandas-docs/stable/timeseries.html#resampling.

    Returns
    -------
    xr.DataArray
      The number of days where value is within [low, high] for each period.
    """
    c = compare(da, ">", low) * compare(da, "<=", high) * 1
    return c.resample(time=freq).sum(dim="time")


def get_daily_events(da: xr.DataArray, da_value: float, operator: str) -> xr.DataArray:
    r"""Return a 0/1 mask when a condition is True or False.

    the function returns 1 where operator(da, da_value) is True
                         0 where operator(da, da_value) is False
                         nan where da is nan

    Parameters
    ----------
    da : xr.DataArray
    da_value : float
    operator : {">", "<", ">=", "<=", "gt", "lt", "ge", "le"}
      Logical operator {>, <, >=, <=, gt, lt, ge, le}. e.g. arr > thresh.

    Returns
    -------
    xr.DataArray
    """
    func = getattr(da, "_binary_op")(get_op(operator))
    events = func(da, da_value) * 1
    events = events.where(~(np.isnan(da)))
    events = events.rename("events")
    return events


def daily_downsampler(da: xr.DataArray, freq: str = "YS") -> xr.DataArray:
    r"""Daily climate data downsampler.

    Parameters
    ----------
    da : xr.DataArray
    freq : str

    Returns
    -------
    xr.DataArray

    Note
    ----

        Usage Example

            grouper = daily_downsampler(da_std, freq='YS')
            x2 = grouper.mean()

            # add time coords to x2 and change dimension tags to time
            time1 = daily_downsampler(da_std.time, freq=freq).first()
            x2.coords['time'] = ('tags', time1.values)
            x2 = x2.swap_dims({'tags': 'time'})
            x2 = x2.sortby('time')
    """
    # generate tags from da.time and freq
    if isinstance(da.time.values[0], np.datetime64):
        years = [f"{y:04d}" for y in da.time.dt.year.values]
        months = [f"{m:02d}" for m in da.time.dt.month.values]
    else:
        # cannot use year, month, season attributes, not available for all calendars ...
        years = [f"{v.year:04d}" for v in da.time.values]
        months = [f"{v.month:02d}" for v in da.time.values]
    seasons = [
        "DJF DJF MAM MAM MAM JJA JJA JJA SON SON SON DJF".split()[int(m) - 1]
        for m in months
    ]

    n_t = da.time.size
    if freq == "YS":
        # year start frequency
        l_tags = years
    elif freq == "MS":
        # month start frequency
        l_tags = [years[i] + months[i] for i in range(n_t)]
    elif freq == "QS-DEC":
        # DJF, MAM, JJA, SON seasons
        # construct tags from list of season+year, increasing year for December
        ys = []
        for i in range(n_t):
            m = months[i]
            s = seasons[i]
            y = years[i]
            if m == "12":
                y = str(int(y) + 1)
            ys.append(y + s)
        l_tags = ys
    else:
        raise RuntimeError(f"Frequency `{freq}` not implemented.")

    # add tags to buffer DataArray
    buffer = da.copy()
    buffer.coords["tags"] = ("time", l_tags)

    # return groupby according to tags
    return buffer.groupby("tags")



# -*- coding: utf-8 -*-
"""Climate indices computation package based on Xarray."""
from xclim.core import units
from xclim.core.options import set_options
from xclim.indicators import ICCLIM, anuclim, atmos, icclim, land, seaIce

__author__ = """Travis Logan"""
__email__ = "logan.travis@ouranos.ca"
__version__ = "0.25.0"



import pytest

from xclim.indicators import build_module


class TestBuildModules:
    def test_nonexistent_process_build_failure(self):
        name = ""
        objs = {"k1": "v1", "k2": "v2"}
        with pytest.raises(AttributeError):
            build_module(name, objs, mode="warn")

    def test_quiet_build_failure(self):
        name = None
        objs = {}
        with pytest.raises(TypeError):
            build_module(name, objs, mode="ignore")

    def test_configured_build_failure(self):
        name = ""
        objs = {"k1": None, "k2": None}
        with pytest.raises(AttributeError):
            build_module(name, objs, mode="bananas")

    def test_loud_build_failure(self):
        name = ""
        objs = {"k1": None, "k2": None}
        with pytest.warns(Warning):
            build_module(name, objs, mode="warn")

    def test_raise_build_failure(self):
        name = ""
        objs = {"k1": None, "k2": None}
        with pytest.raises(NotImplementedError):
            build_module(name, objs, mode="raise")


class TestICCLIM:
    def test_exists(self):
        from xclim import icclim

        assert getattr(icclim, "TG", None) is not None


class TestANUCLIM:
    def test_exists(self):
        from xclim import anuclim

        assert getattr(anuclim, "P1_AnnMeanTemp", None) is not None
        assert getattr(anuclim, "P19_PrecipColdestQuarter", None) is not None
